import numpy as np
import tensorflow as tf
from sklearn.base import BaseEstimator, ClassifierMixin
from ..utils import fuzzy_hinge, fuzzy_cross_entropy

class GRMLinearClassifier(BaseEstimator, ClassifierMixin):
    '''
    A class to perform classification for weakly supervised data, based on the Generalized Risk Minimization Paradigm
    applied to linear classifiers (either, logistic regression or linear SVM).
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel

    Parameters
    ----------
    :param loss: The loss function to optimize
    :type loss: str or callable, default 'logistic'

    :param max_epochs: The number of epochs to train the model
    :type max_epochs: int, default 100

    :param optimizer: The optimizer algorithm
    :type optimizer: str or callable, default 'sgd'

    :param regularizer: The type of regularization to apply
    :type regularizer: str or callable, default None

    :param l1: The regularization coefficient for l1 regularization, used only if regularizer='l1'
    :type l1: float, default 0.01

    :param l2: The regularization coefficient for l2 regularization, used only if regularizer='l2'
    :type l2: float, default 0.01

    :param batch_size: Size of the mini-batches
    :type batch_size: int, default 32



    Attributes
    ----------
    :ivar regularization: The regularization method instance
    :vartype y: tf.keras.regularizers.Regularizer

    :ivar model: The fitted linear model
    :vartype model: tf.keras.Model

    :ivar loss_funct: The loss function callable
    :vartype loss_funct: callable

    :ivar __n_classes: The number of unique classes in y
    :vartype __n_classes: int

    :ivar __classes: The unique classes in y
    :vartype __classes: list of int
    '''
    
    def __init__(self, loss="logistic", max_epochs = 100, optimizer='sgd', random_state=None,
                 regularizer=None, l1=0.01, l2=0.01, batch_size=32):
        self.loss = loss
        self.max_epochs = max_epochs
        self.optimizer = optimizer
        self.random_state = random_state
        self.regularizer = regularizer
        self.l1 = l1
        self.l2 = l2
        self.batch_size = batch_size
        
    def fit(self, X, y):
        """
        Fit the GRMLinearClassifier model
        """
        tf.config.run_functions_eagerly(True)
        np_state = np.random.get_state()

        if self.random_state is not None:
            np.random.seed(self.random_state)
            tf.random.set_seed(self.random_state)
        
        self.loss_funct = None
        self.model = None
        self.__X = X
        self.__y = y

        self.__n_classes = self.__y[0].n_classes

        self.__y_probs = np.empty((self.__X.shape[0], self.__n_classes))

        for i in range(len(self.__y)):
            self.__y_probs[i] = self.__y[i].to_probs()
        
        self.__n_features = self.__X.shape[1]
        
        self.regularization = None
        if self.regularizer == "l1":
            self.regularization = tf.keras.regularizers.l1(self.l1)
        elif self.regularizer == "l2":
            self.regularization = tf.keras.regularizers.l1(self.l2)
        elif self.regularizer == "l1_l2":
            self.regularization = tf.keras.regularizers.l1_l2(self.l1, self.l2)
        
        self.model = None
        if self.loss == "logistic":
            self.loss_funct = fuzzy_cross_entropy
            self.model = tf.keras.models.Sequential([
                tf.keras.layers.Dense(self.__n_classes,
                                      activation='softmax',
                                      kernel_regularizer=self.regularization,
                                      input_shape=[self.__n_features])
            ])
        elif self.loss == "hinge":
            self.loss_funct = fuzzy_hinge
            self.model = tf.keras.models.Sequential([
                tf.keras.layers.Dense(self.__n_classes,
                                      activation='linear',
                                      kernel_regularizer=self.regularization,
                                      input_shape=[self.__n_features])
            ])
        print(tf.executing_eagerly())
        self.model.compile(loss = self.loss_funct,
                           optimizer = self.optimizer)
        self.model.fit(self.__X, self.__y_probs, batch_size = self.batch_size, epochs = self.max_epochs, verbose=0)

        if self.random_state is not None:
            np.random.set_state(np_state)
            tf.random.set_seed(np.random.randint(10000000))

        tf.config.run_functions_eagerly(False)
        
        return self
        
    def predict_proba(self, X):
        """
        Returns probability distributions for the given X
        """
        return self.model.predict(X)

    def predict(self, X):
        """
        Returns predictions for the given X
        """
        y_pred = np.zeros(X.shape[0])
        preds = self.predict_proba(X)
        for i in range(len(y_pred)):
            y_pred[i] = np.argmax(np.add.reduce(preds[i,:]))
        return y_pred


import pytest
import numpy as np
from ...classification import GRMLinearClassifier
from sklearn.datasets import load_iris
from ...data_representation import *

from sklearn.neighbors import KNeighborsClassifier

@pytest.fixture
def dataset():
    X, y = load_iris(return_X_y=True)
    clf = KNeighborsClassifier(n_neighbors=10)
    clf.fit(X,y)
    y_fuzzy = clf.predict_proba(X)
    y_fuzzy = y_fuzzy/np.max(y_fuzzy, axis=1)[:, np.newaxis]
    y_soft = np.empty(y_fuzzy.shape[0], dtype=DiscreteFuzzyLabel)
    for i in range(y_fuzzy.shape[0]):
        y_soft[i] = DiscreteFuzzyLabel(y_fuzzy[i], 3)
    return (X, y, y_soft)

def test_grm_linear(dataset):
    X, y_true, y = dataset[0], dataset[1], dataset[2]
    clf = GRMLinearClassifier(max_epochs=1)
    clf.fit(X, y)
    y_pred = clf.predict(X)
    assert True



import numpy as np
import itertools
from sklearn.base import BaseEstimator, TransformerMixin, clone
from ..classification import WeaklySupervisedKNeighborsClassifier, WeaklySupervisedRadiusClassifier
from sklearn.metrics import accuracy_score


class RoughSetSelector(TransformerMixin, BaseEstimator):
    """
    A class to perform Feature Selection based on Rough Sets by searching for reducts [1].
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel.
    Supports both discrete (using Pawlak Rough Sets) and continuous (using Neighborhood Rough Sets) datasets.

    Parameters
    ----------
    :param search_strategy: The search strategy to be used. 'approximate' is similar to RFE, having complexity O(n^2). 'brute' is a brute-force search strategy, all possible combinations of features are evaluated, with complexity O(2^n)
    :type search_strategy: {'approximate', 'brute'}, default='approximate'

    :param epsilon: The approximation factor. Should be a number between 0.0 and 1.0 (excluded)
    :type epsilon: float, default=0.0

    :param n_iters: Number of iterations to be used when search_strategy='approximate'. Not used if search_strategy='brute'
    :type n_iters: int, default=100

    :param method: The method used to compute the fitness. If 'lambda', then the algorithm solves a single objective optimization problem
        If 'conservative' solve a 2-objectives problem
    :type method: {'lambda', 'conservative'}, default='conservative'

    :param l: Lambda interpolation factor. Only used if method='lambda'
    :type l: float in [0,1], default=0.5

    :param discrete: Whether the input X is discrete or not. If discrete=True then use equivalence-based
        (i.e. Pawlak) Rough Sets. If discrete=False use neighborhood-based Rough Sets.
    :type discrete: bool, default=True
        
    :param metric: Metric to be used with neighborhood-based Rough Sets. Only used if discrete=False. If discrete=True, then metric="hamming"
    :type metric: string or function, default='minkowski'
        
    :param neighborhood: Type of neighborhood-based Rough Sets to be used. If neighborhood='delta', then
        use delta-neighborhood Rough Sets: all neigbhors with distance <= radius are selected.
        If neighborhood='nearest', then use k-nearest-neighbors Rough Sets: only the k
        nearest neighbors are selected. Only used if discrete=False
    :type neighborhood: {'delta', 'nearest'}, default='nearest'

    :param n_neighbors: Number of nearest neighbors to select. Only used if discrete=False and
        neighborhood='nearest'
    :type n_neighbors: int, default=3

    :param radius: Radius to select neighbors. Only used if discrete=False and neighborhood='delta'
    :type radius: float, default=1.0

    :param random_state: Randomization seed. Used only if search_strategy='approximate'
    :type random_state: int, default=None

    Attributes
    ----------
    :ivar n_classes_: The number of unique classes in y
    :vartype n_classes: int

    :ivar reducts_: The list of minimal reducts. If search_strategy='approximate', reducts always contains at most a single set of features for each membership degree. If search_strategy='brute', reducts contains the list of all minimal reducts.
    :vartype reducts_: list

    :ivar reducts_poss_: The list of membership values of the minimal reducts.
    :vartype reducts_: list
    """
    def __init__(self, search_strategy='approximate', epsilon = 0.0,
                 n_iters = 100, method='conservative', l = 0.5, discrete = True, metric='minkowski',
                 neighborhood='nearest', n_neighbors=3, radius=1.0, random_state = None):
        self.search_strategy = search_strategy
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.random_state = random_state
        self.discrete = discrete
        self.n_neighbors = n_neighbors
        self.radius = radius
        self.neighborhood = neighborhood
        self.metric = metric
        self.method = method
        self.l = l
        
    def fit(self, X, y):
        """
        Fit the RoughSetSelector model
        """
        state = np.random.get_state()
        if not (self.random_state is None):
            np.random.seed(self.random_state)


        self.__X = X
        self.__dim = self.__X.shape[1]
        self.reducts_ = []
        self.reducts_poss_ = []
        self.__y = np.array(y)
        self.n_classes_ = self.__y[0].n_classes

        self.__y_values = np.zeros((self.__X.shape[0], self.n_classes_))
        for i in range(self.__X.shape[0]):
            self.__y_values[i,:] = self.__y[i].classes
        self.__y_values = np.unique(self.__y_values)[::-1]
        self.__y_values = self.__y_values[self.__y_values > 0.0]
            
        self.__nn = None
        if self.discrete:
            self.__nn = WeaklySupervisedRadiusClassifier(radius = 0.0,
                                          metric = 'hamming')
        elif self.neighborhood == 'delta':
            self.__nn = WeaklySupervisedRadiusClassifier(radius=self.radius,
                                           metric=self.metric)
        elif self.neighborhood == 'nearest':
            self.__nn = WeaklySupervisedKNeighborsClassifier(k = self.n_neighbors,
                                      metric = self.metric)
        else:
            raise ValueError("%s is not an allowed value for 'neighborhood' parameter" % self.neighborhood)

        for v in self.__y_values:
            y_cut = []
            for i in range(self.__X.shape[0]):
                y_cut.append(self.__y[i].get_cut(v))
            y_cut = np.array(y_cut)

            self.__nn_all = clone(self.__nn)
            self.__nn_all.fit(self.__X, y_cut)
        
            temp_reducts = []
            if self.search_strategy == 'approximate':
                temp_reducts = [self.__find_approx_reducts(y_cut)]
            elif self.search_strategy == 'brute':
                temp_reducts = self.__find_reducts(y_cut)
            else:
                raise ValueError("%s is not an allowed value for 'search_strategy' parameter." % self.search_strategy)

            if len(self.reducts_) == 0:
                self.reducts_ = temp_reducts
                self.reducts_poss_ = [v]*len(self.reducts_)
            else:
                for temp_red in temp_reducts:
                    check = True
                    for i in range(len(self.reducts_)):
                        if self.method == 'conservative':
                            if len(temp_red) >= len(self.reducts_[i]):
                                check = False
                                break
                        if self.method == 'lambda':
                            red_l = (1 - self.l)*self.reducts_poss_[i]  - self.l*len(self.reducts_[i])/self.__dim
                            temp_red_l = (1 - self.l)*v - self.l*len(temp_red)/self.__dim
                            if temp_red_l > red_l:
                                check = False
                                break
                    if check:
                        self.reducts_.append(temp_red)
                        self.reducts_poss_.append(v)

            if not (self.random_state is None):
                np.random.set_state(state)
        return self
    
    def transform(self, X, y=None):
        """
        Transform the data (only X, y is ignored) selecting a reduct at random
        """
        state = np.random.get_state()
        if not (self.random_state is None):
            np.random.seed(self.random_state)
            
        feature_set = np.random.choice(len(self.reducts_))

        if not (self.random_state is None):
            np.random.set_state(state)

        return X[:, self.reducts_[feature_set]]
    
    def fit_transform(self, X, y):
        """
        Fit and then transform data
        """
        self.fit(X, y)
        return self.transform(X, y)
    
    def __find_reducts(self, y_cut):
        reducts = []
        min_len = self.__dim
        for k in range(self.__dim, 0, -1):
            for i in itertools.combinations(range(self.__dim), k):
                red = np.array(list(i))
                temp_X = self.__X[:, red].copy()
                self.__nn.fit(temp_X, y_cut)

                accuracy = accuracy_score(self.__nn_all.predict(self.__X), self.__nn.predict(temp_X))
                check = (accuracy >= (1 - self.epsilon))
               
                if check:
                    if len(red) < min_len:
                        min_len = len(red)
                        reducts = [red]
                    elif len(red) == min_len:
                        reducts.append(red)
        return np.array(reducts)
    
    def __find_approx_reducts(self, y_cut):
        total_red = range(self.__dim)

        for iter in range(self.n_iters):      
            red = range(self.__dim)
            check = True
            while check == True and len(red) > 1:
                check = False
                to_keep = red
                combs = itertools.combinations(red, len(red) - 1)
                combs = np.random.permutation(list(combs))
                for i in combs:

                    temp_red = np.array(list(i))
                    temp_X = self.__X[:, temp_red].copy()
                    self.__nn.fit(temp_X, y_cut)

                    accuracy = accuracy_score(self.__nn_all.predict(self.__X), self.__nn.predict(temp_X))
                    check_acc = (accuracy >= (1 - self.epsilon))
                
                    if check_acc:
                        to_keep = temp_red
                    
                if len(to_keep) < len(red):
                    red = to_keep
                    check = True
            if len(red) < len(total_red):
                total_red = red
        return total_red


import pytest
import numpy as np
from ...data_representation import *


@pytest.fixture
def discrete_fuzzy_label():
    '''Returns a fuzzy label with 5 classes'''
    return DiscreteFuzzyLabel(np.array([0.0, 0.6, 1.0, 1.0, 0.8]), 5)

@pytest.fixture
def discrete_set_label():
    '''Returns a superset label with 5 classes'''
    return DiscreteSetLabel(np.array([0.0, 0.0, 1.0, 1.0, 1.0]), 5)

@pytest.fixture
def interval_label():
    '''Returns a default interval label'''
    return IntervalLabel(0, 3.5)


###Discrete Labels
def test_fuzzy_label_from_dict(discrete_fuzzy_label):
    assert DiscreteFuzzyLabel({1: 0.6, 2: 1.0, 3: 1.0, 4: 0.8}, 5) == discrete_fuzzy_label

def test_set_label_from_list(discrete_set_label):
    assert DiscreteFuzzyLabel([2,3,4], 5) == discrete_set_label

def test_alpha_cut_discrete(discrete_fuzzy_label, discrete_set_label):
    assert discrete_fuzzy_label.get_cut(0.7) == discrete_set_label

def test_membership_discrete(discrete_fuzzy_label):
    assert discrete_fuzzy_label[1] == 0.6

def test_sample_value_discrete(discrete_fuzzy_label):
    check = True
    for i in range(100):
        check = (check) | (discrete_fuzzy_label.sample_value != 0)
    assert check != False


### Interval Labels

def test_membership_interval_in(interval_label):
    assert interval_label[2.2] == 1.0

def test_membership_interval_border(interval_label):
    assert interval_label[3.5] == 1.0

def test_membership_interval_out(interval_label):
    assert interval_label[-1] == 0.0

def test_alpha_cut_interval(interval_label):
    assert interval_label == interval_label.get_cut(0.5)

def test_sample_value_interval(interval_label):
    check = True
    for i in range(100):
        check = (check) or ((interval_label.sample_value() >= interval_label.lower) and (interval_label.sample_value() <= interval_label.upper))
    assert check != False



from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
from ..data_representation import *

class DiscreteEstimatorSmoother(TransformerMixin, BaseEstimator):
    '''
    A class to transform a supervised learning problem into a weakly supervised one, based on the output of a classifier. It
    currently supports transformation to superset and fuzzy label learning. Note that DiscreteEstimatorSmoother does not implement
    the transform method: therefore, usage should be based on calling fit_transform.

    Parameters
    ----------
    :param estimator: Base estimator objects to be fitted. Should support predict_proba
    :type estimator: estimator class

    :param type: Type of weakly supervised labels to transform into
    :type n_estimators: {'set', 'fuzzy'}, default=set

    :param epsilon: Parameter to select the minimum allowed label degree. Only used when type == 'set'. Should be between 0 and 1
    :type epsilon: float, default=1.0

    Attributes
    ----------
    :ivar n_classes_: The number of unique classes in y
    :vartype n_classes: int
    '''
    def __init__(self, estimator, type='set', epsilon=1.0):
        self.estimator=estimator
        self.type=type
        self.epsilon=epsilon

    def fit(self, X, y):
        self.__base_type = None
        if self.type == 'set':
            self.__base_type = DiscreteSetLabel
        elif self.type == 'fuzzy':
            self.__base_type = DiscreteFuzzyLabel
        else:
            raise ValueError("%s is not an allowed value for 'type' parameter" % self.type)

        self.n_classes_ = len(np.unique(y))
        self.estimator.fit(X,y)
        return self

    def fit_transform(self, X, y):
        self.fit(X, y)
        y_fuzzy = self.estimator.predict_proba(X)
        y_fuzzy = y_fuzzy/np.max(y_fuzzy, axis=1)[:, np.newaxis]
        y_soft = np.empty(y_fuzzy.shape[0], dtype=self.__base_type)

        for i in range(y_fuzzy.shape[0]):
            if self.type == 'set':
                y_fuzzy[i] = (y_fuzzy[i] > self.epsilon)
            y_soft[i] = self.__base_type(y_fuzzy[i], self.n_classes_)
        
        return y_soft

class DiscreteRandomSmoother(TransformerMixin, BaseEstimator):
    '''
    A class to transform a supervised learning problem into a weakly supervised one, based on random sampling. It
    currently supports transformation to superset and fuzzy label learning. Note that DiscreteRandomSmoother does not implement
    the transform method: therefore, usage should be based on calling fit_transform.

    Parameters
    ----------
    :param p_err: The probability to include any single wrong label in a sample draw. Should be between 0 and 1.
    :type p_err: float, default=0.1

    :param p_incl: The probability to include the correct label in a sample draw. Should be between 0 and 1.
    :type p_incl: float, default=1.0

    :param prob_ranges: Array of possible membership degrees to be sampled. If not None, overrides both p_err and p_incl
    :type prob_ranges: enumerable of float, default=None

    :param type: Type of weakly supervised labels to transform into
    :type type: {'set', 'fuzzy'}, default=set

    :param samples: Number of samples to be generated
    :type samples: int, default=100

    :param epsilon: Parameter to select the minimum allowed label degree. Only used when type == 'set'. Should be between 0 and 1
    :type epsilon: float, default=1.0

    Attributes
    ----------
    :ivar n_classes_: The number of unique classes in y
    :vartype n_classes: int
    '''
    def __init__(self, p_err = 0.1, p_incl = 1.0, prob_ranges=None, type='set', samples=100, epsilon=0.0):
        self.p_err = p_err
        self.p_incl = p_incl
        self.type = type
        self.samples = samples
        self.epsilon = epsilon
        self.prob_ranges = prob_ranges

    def fit(self, X, y):
        self.__base_type = None
        if self.type == 'set':
            self.__base_type = DiscreteSetLabel
        elif self.type == 'fuzzy':
            self.__base_type = DiscreteFuzzyLabel
        else:
            raise ValueError("%s is not an allowed value for 'type' parameter" % self.type)
        
        self.n_classes_ = len(np.unique(y))
        return self

    def fit_transform(self, X, y):
        self.fit(X,y)

        y_soft = np.empty(y.shape[0], dtype=self.__base_type)
        y_temp = np.zeros((y.shape[0], self.n_classes_))

        if self.prob_ranges is not None:
            y_temp = np.random.choice(self.prob_ranges, size=(y.shape[0], self.n_classes_))

        for i in range(X.shape[0]):
            if self.prob_ranges is None:
                y_temp[i, y[i]] = 1.0
                for k in range(self.n_classes_):
                    for s in range(self.samples):
                        if k == y[i]:
                            y_temp[i, k] += np.random.choice([0,1], p=[1-self.p_incl, self.p_incl])
                        else:
                            y_temp[i, k] += np.random.choice([0,1], p=[1-self.p_err, self.p_err])
                
                y_temp[i] /= self.samples
                y_temp[i] = y_temp[i]/np.max(y_temp[i])
                if self.type == 'set':
                    y_temp[i] = (y_temp[i] > self.epsilon)*1.0
                
            y_soft[i] = self.__base_type(y_temp[i], self.n_classes_)

        return y_soft




import pytest
import numpy as np
from ...feature_selection import GeneticRoughSetSelector
from ...data_representation import *
from sklearn.datasets import load_iris
from sklearn.neighbors import KNeighborsClassifier

@pytest.fixture
def dataset():
    X, y = load_iris(return_X_y=True)
    clf = KNeighborsClassifier(n_neighbors=10)
    clf.fit(X,y)
    y_fuzzy = clf.predict_proba(X)
    y_fuzzy = y_fuzzy/np.max(y_fuzzy, axis=1)[:, np.newaxis]
    y_soft = np.empty(y_fuzzy.shape[0], dtype=DiscreteFuzzyLabel)
    for i in range(y_fuzzy.shape[0]):
        y_soft[i] = DiscreteFuzzyLabel(y_fuzzy[i], 3)
    return (X, y, y_soft)

def test_simple(dataset):
    rss = GeneticRoughSetSelector(n_iters=10, epsilon=0.1, random_state=0)
    X, y = dataset[0], dataset[2]
    rss.fit(X,y)
    rss.transform(X)
    assert True

def test_random_state(dataset):
    rss = GeneticRoughSetSelector(n_iters=10, epsilon=0.1, random_state=0)
    X, y = dataset[0], dataset[2]
    res1 = rss.fit(X,y).best_features_[0]
    res2 = rss.fit(X,y).best_features_[0]
    assert np.array_equal(res1, res2) 

def test_transform(dataset):
    rss = GeneticRoughSetSelector(n_iters=10, epsilon=0.1, random_state=0)
    X, y = dataset[0], dataset[2]
    res1 = rss.fit_transform(X,y)
    res2 = rss.fit_transform(X,y)
    assert np.array_equal(res1, res2)


from ._roughset import RoughSetSelector
from ._geneticroughset import GeneticRoughSetSelector
from ._delin import DELIN

__all__ = [
    "RoughSetSelector",
    "GeneticRoughSetSelector",
    "DELIN"
]


from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.linear_model import LogisticRegression
import numpy as np

from scikit_weak.classification._relaxation import LabelRelaxationNNClassifier


class PseudoLabelsClassifier(BaseEstimator, ClassifierMixin):
    '''
      A class to perform classification for weakly supervised data, based on the pseudo-labels strategy.
      The y input to the fit method should be given as an iterable of GenericWeakLabel

      Parameters
      ----------
      :param estimator: Base estimator objects to be fitted. Should support predict and predict_proba
      :type estimator: estimator class, default=LogisticRegression

      :param n_restarts: The number of restarts
      :type n_restarts: int, default = 5

      :param n_iterations: The number of iterations for fitting
      :type n_iterations: int, default=10

      :param threshold: The threshold for pseudo-label selection
      :type threshold: float, default=0.5

      :param random_state: Random seed
      :type random_state: int, default=None

      Attributes
      ----------

      :ivar estimator: The last fitted estimator
      :vartype estimator: estimator

      :ivar __n_classes: The number of unique classes in y
      :vartype __n_classes: int

      :ivar __classes: The unique classes in y
      :vartype __classes: list of int
      '''

    def __init__(self, estimator=LogisticRegression(), n_iterations=10, n_restarts=5, threshold=0.5, random_state=None):
        self.n_iterations = n_iterations
        self.n_restarts = n_restarts
        self.random_state = random_state
        self.estimator = estimator
        self.threshold = threshold

    def fit(self, X, y):
        """
            Fit the PseudoLabelsClassifier model
        """
        state = np.random.get_state()
        if not (self.random_state is None):
            np.random.seed(self.random_state)

        self.__X = X
        self.__y = np.array(y)

        self.__n_classes = self.__y[0].n_classes
        self.__classes = range(self.__n_classes)

        acc = 0

        for r in range(self.n_restarts):

            self.__fix = [False] * len(self.__y)
            self.__imputed_y = np.random.randint(0, self.__n_classes, self.__y.shape)

            temp_est = clone(self.estimator)

            for i in range(self.n_iterations):
                self.__sample_labels()
                temp_est.fit(self.__X, self.__imputed_y)

                y_proba = temp_est.predict_proba(self.__X)
                for t in range(y_proba.shape[0]):
                    if y_proba[t, self.__imputed_y[t]] > self.threshold:
                        self.__fix[t] = True

            y_pred = temp_est.predict(self.__X)
            temp_acc = 0
            for t in range(y_pred.shape[0]):
                temp_acc += self.__y[t][y_pred[t]]
            if temp_acc > acc:
                self.estimator = temp_est

        if not (self.random_state is None):
            np.random.set_state(state)
        return self

    def __sample_labels(self):
        for i in range(self.__y.shape[0]):
            if self.__fix[i] == False:
                self.__imputed_y[i] = self.__y[i].sample_value()

    def predict(self, X):
        """
            Returns predictions for the given X
        """
        return self.estimator.predict(X)

    def predict_proba(self, X):
        """
            Returns probability distributions for the given X
        """
        return self.estimator.predict_proba(X)

    def fit_predict(self, X, y):
        self.fit(X, y)
        return self.predict(X)


class CSSLClassifier(BaseEstimator, ClassifierMixin):
    '''
      A credal self-supervised learning classifiers as described in "Credal Self-Supervised Learning" by
        Julian Lienen and Eyke Huellermeier, NeurIPS 2021. The model iteratively refines its beliefs about the true
        conditional class probability distribution for weakly labeled instances by maintaining credal sets induced by
        possibility distributions.

      Parameters
      ----------
      :param estimator: Base estimator objects to be fitted. Should support predict and predict_proba
      :type estimator: estimator class, default=LabelRelaxationNNClassifier

      :param n_iterations: The number of iterations for fitting
      :type n_iterations: int, default=10

      :param random_state: Random seed
      :type random_state: int, default=None

      :param p_data: Optional prior probability distribution of the class frequencies
      :type p_data: np.ndarray, default=None

      :param p_hist_buffer_size: Buffer size of the model prediction history
      :type p_hist_buffer_size: int, default=64

      Attributes
      ----------

      :ivar estimator: The last fitted estimator
      :vartype estimator: estimator

      :ivar _n_classes: The number of unique classes in y
      :vartype _n_classes: int
      '''

    def __init__(self, estimator=LabelRelaxationNNClassifier(provide_alphas=True), n_iterations=10,
                 n_classes=10, random_state=None, p_data=None, p_hist_buffer_size=64):
        self.n_iterations = n_iterations
        self.random_state = random_state
        self.estimator = estimator
        self.estimator.n_classes = n_classes

        self.n_classes = n_classes
        if p_data is not None:
            self.p_data = p_data
        else:
            self.p_data = np.ones(n_classes) / n_classes
        self.p_hist = np.ones([p_hist_buffer_size, self.n_classes]) / self.n_classes

    def fit(self, X, y):
        """
        y is list of discrete fuzzy labels
        """
        state = np.random.get_state()
        if not (self.random_state is None):
            np.random.seed(self.random_state)

        self.__X = X

        # Prepare targets and alphas
        self.__y = []
        self.__alphas = []
        for y_val in y:
            if y_val.is_agnostic():
                self.__y.append(-1)
                self.__alphas.append(1.)
            else:
                self.__y.append(np.argmax(y_val.classes))
                self.__alphas.append(0.)
        self.__y = np.array(self.__y)
        self.__alphas = np.array(self.__alphas)

        # Iteratively refine alpha values until there is no imprecisiation left or number of iterations is exhausted
        for i in range(self.n_iterations):
            # Fit model on current alphas
            train_mask = self.__y != -1
            train_X = self.__X[train_mask]
            train_y = self.__y[train_mask]
            train_alphas = self.__alphas[train_mask]

            self.estimator = self.estimator.fit(train_X, (train_y, train_alphas))

            # Update alphas
            update_mask = self.__alphas < 1.
            preds = self.estimator.predict(self.__X[update_mask])

            p_model = self.p_hist.mean(axis=0)
            p_model /= p_model.sum()

            norm_preds = preds * (self.p_data + 1e-7) / (p_model + 1e-7)
            norm_preds /= norm_preds.sum()

            self.__y[update_mask] = np.argmax(norm_preds, axis=-1)
            self.__alphas[update_mask] = 1. - np.max(norm_preds, axis=-1)

            # Update p_hist
            self.p_hist = np.concatenate((self.p_hist[1:], np.expand_dims(norm_preds.mean(axis=0), axis=0)), axis=0)

        if not (self.random_state is None):
            np.random.set_state(state)
        return self

    def predict(self, X):
        """
            Returns predictions for the given X
        """
        return self.estimator.predict(X)

    def predict_proba(self, X):
        """
            Returns probability distributions for the given X
        """
        return self.estimator.predict_proba(X)

    def fit_predict(self, X, y):
        self.fit(X, y)
        return self.predict(X)



from ._utils import bet_entropy, oau_entropy
from ._smoothers import DiscreteEstimatorSmoother, DiscreteRandomSmoother
from ._losses import fuzzy_cross_entropy, fuzzy_hinge, fuzzy_loss

__all__ = [
    'bet_entropy',
    'oau_entropy',
    'DiscreteEstimatorSmoother',
    'DiscreteRandomSmoother',
    'fuzzy_cross_entropy',
    'fuzzy_hinge',
    'fuzzy_loss'
]


import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.neighbors import KNeighborsClassifier, RadiusNeighborsClassifier
from sklearn.metrics import accuracy_score
from abc import ABC, abstractmethod
from sklearn.base import clone


class FitnessVal(ABC):
    @abstractmethod
    def __gt__(self, other):
        pass

    @abstractmethod
    def __eq__(self, other):
        pass
    
class SimpleVal(FitnessVal):
    def __init__(self, val):
        self.val = val

    def __gt__(self, other):
        return  self.val > other.val

    def __eq__(self, other):
        return self.val == other.val

class SimplePair(FitnessVal):
    def __init__(self, reduct_size, poss_value):
        self.reduct_size = reduct_size
        self.poss_value = poss_value

    def __gt__(self, other):
        check = False

        if (self.reduct_size < other.reduct_size) and (self.poss_value >= other.poss_value):
            check = True
        elif (self.reduct_size <= other.reduct_size) and (self.poss_value > other.poss_value):
            check = True
        
        return check

    def __eq__(self, other):
        return not((self > other) or (other > self))

class ArrayPair(FitnessVal):
    def __init__(self, reduct_size, poss_array):
        self.reduct_size = reduct_size
        self.poss_array = poss_array

    def __gt__(self, other):
        check = False

        weak = np.all(self.poss_array >= other.poss_array)
        strong = np.any(self.poss_array > other.poss_array)

        if (self.reduct_size < other.reduct_size) and (weak):
            check = True
        elif (self.reduct_size == other.reduct_size) and (weak and strong):
            check =True

        return check

    def __eq__(self, other):
        return not((self > other) or (other > self))


class GeneticRoughSetSelector(TransformerMixin, BaseEstimator):
    """
    A class to perform Rough Set-based feature selection, by searching for reducts, using Genetic Algorithms.
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel.
    Supports both discrete (using Pawlak Rough Sets) and continuous (using Neighborhood Rough Sets) datasets.

    Parameters
    ----------
    :param epsilon: The approximation factor. Should be a number between 0.0 and 1.0 (excluded)
    :type epsilon: float, default=0.0

    :param method: The method used to compute the fitness. If 'lambda', then the algorithm solves a single objective optimization problem
        If 'conservative' or 'dominance' solves a multiple objective optimization problem: in particular, if 'conservative' a 2-objectives problem
        and if 'dominance' (n+1)-objectives problem, where n is the number of instances
    :type method: {'lambda', 'conservative', 'dominance'}, default='conservative'

    :param discrete: Whether the input X is discrete or not. If discrete=True then use equivalence-based
        (i.e. Pawlak) Rough Sets. If discrete=False use neighborhood-based Rough Sets.
    :type discrete: bool, default=True
        
    :param l: Lambda interpolation factor. Only used if method='lambda'
    :type l: float in [0,1], default=0.5

    :param tournament_size: Proportion of population to select from in tournament selection.
    :type tournament_size: float in [0,1], default=0.1

    :param p_mutate: Proability of point mutation
    :type p_mutate: float in [0,1], default=0.1

    :param metric: Metric to be used with neighborhood-based Rough Sets. Only used if discrete=False. If discrete=True, then metric="hamming"
    :type metric: string or function, default='minkowski'
        
    :param neighborhood: Type of neighborhood-based Rough Sets to be used. If neighborhood='delta', then
        use delta-neighborhood Rough Sets: all neigbhors with distance <= radius are selected.
        If neighborhood='nearest', then use k-nearest-neighbors Rough Sets: only the k
        nearest neighbors are selected. Only used if discrete=False
    :type neighborhood: {'delta', 'nearest'}, default='nearest'

    :param n_neighbors: Number of nearest neighbors to select. Only used if discrete=False and
        neighborhood='nearest'
    :type n_neighbors: int, default=3

    :param radius: Radius to select neighbors. Only used if discrete=False and neighborhood='delta'
    :type radius: float, default=1.0

    :param random_state: Randomization seed. Used only if search_strategy='approximate'
    :type random_state: int, default=None

    :param population_size: Size of the population for the Genetic Algorithm
    :type population_size: int, default=100

    :param n_iters: Number of generations for the Genetic Algorithm
    :type n_iters: int, default=100

    Attributes
    ----------
    :ivar n_classes: The number of unique classes in y
    :vartype n_classes_: int

    :ivar best_features_: The unique most fit feature sets.
    :vartype best_features_: ndarray

    :ivar best_targets_: The disambiguated targets corresponing to the most fit feature sets. Can be used for transductive learning or training a downstream model.
    :vartype best_target_: ndarray
    """
    def __init__(self, epsilon = 0.0, method = 'conservative',
                 discrete = False, l = 0.5, tournament_size = 0.1, p_mutate=0.1,
                 metric='minkowski', neighborhood='nearest', n_neighbors=3, radius=1.0,
                 random_state = None, population_size = 100, n_iters = 100):
        self.epsilon = epsilon
        self.n_iters = n_iters
        self.random_state = random_state
        self.discrete = discrete
        self.n_neighbors = n_neighbors
        self.radius = radius
        self.neighborhood = neighborhood
        self.metric = metric
        self.population_size = population_size
        self.method = method
        self.l = l
        self.tournament_size = tournament_size
        self.p_mutate = p_mutate

    def _initialize(self):
        self.__population_features = np.random.choice(a=[False, True], size=(self.population_size, self.__dim))
        self.__population_targets = np.empty((self.population_size, self.__n_instances), dtype=self.__basic_type)
        self._true_targets = np.empty((self.population_size, self.__n_instances), dtype=object)
        for i in range(self.population_size):
            self.__population_targets[i, :] = [self.y[j].sample_value() for j in range(self.__n_instances)]
            self._true_targets[i, :] = self.y

    def __breed(self):
        i = 0
        features_temp = self.__population_features.copy()
        targets_temp = self.__population_targets.copy()
        true_targets_temp = self._true_targets.copy()
        for c in range(int(self.population_size/2)):
            parent_1 = self.__select()
            parent_2 = self.__select()

            child_1, child_2 = self.__crossover(parent_1, parent_2)
            
            child_1 = self.__mutate(child_1)
            child_2 = self.__mutate(child_2)

            features_temp[i,:] = child_1[0]
            targets_temp[i, :] = child_1[1]
            true_targets_temp[i, :] = child_1[2]

            features_temp[i+1,:] = child_2[0]
            targets_temp[i+1, :] = child_2[1]
            true_targets_temp[i+1, :] = child_2[2]

            i+=2

        self.__population_features = features_temp
        self.__population_targets = targets_temp
        self._true_targets = true_targets_temp

    def __select(self):
        indices = np.random.choice(self.population_size,
                                   int(self.population_size*self.tournament_size),
                                   replace=False)

        ind = indices[0]
        for i in indices:
            if self.__fitnesses[i] > self.__fitnesses[ind]:
                ind = i
        
        return ind

    def __mutate(self, child):
        features = child[0]
        targets = child[1]

        draw = np.random.choice([True, False],
                                self.__dim,
                                p=[self.p_mutate, 1 - self.p_mutate])
        features = features ^ draw

        draw = np.random.choice([True, False],
                                self.__n_instances,
                                p=[self.p_mutate, 1 - self.p_mutate])

        indices = np.array(range(self.__n_instances))[draw]
        for i in indices:
            targets[i] = child[2][i].sample_value()

        return (features, targets, child[2])

    def __crossover(self, a, b):
        features_split_ind = np.random.choice(self.__dim)
        target_split_ind = np.random.choice(self.__n_instances)

        child_1_features = self.__population_features[a].copy()
        child_1_features[features_split_ind:] = self.__population_features[b, features_split_ind:]
        child_1_targets = self.__population_targets[a].copy()
        child_1_targets[target_split_ind:] = self.__population_targets[b, target_split_ind:]
        child_1_true_targets = self._true_targets[a].copy()
        child_1_true_targets[target_split_ind:] = self._true_targets[b, target_split_ind:]

        child_2_features = self.__population_features[b].copy()
        child_2_features[features_split_ind:] = self.__population_features[a, features_split_ind:]
        child_2_targets = self.__population_targets[b].copy()
        child_2_targets[target_split_ind:] = self.__population_targets[a, target_split_ind:]
        child_2_true_targets = self._true_targets[b].copy()
        child_2_true_targets[target_split_ind:] = self._true_targets[a, target_split_ind:]

        child_1 = (child_1_features, child_1_targets, child_1_true_targets)
        child_2 = (child_2_features, child_2_targets, child_2_true_targets)
        return child_1, child_2
                

    def __compute_fitness(self, p):
        features = self.__population_features[p,:]
        
        temp_X = self.X[:, features]
        temp_y = self.__population_targets[p]

        reduct_val = self.__evaluate_reduct(temp_X, temp_y)
        poss_val = self.__evaluate_poss(temp_y)
        
        if self.method == 'conservative':
            return SimplePair(reduct_val, poss_val)
        elif self.method == 'lambda':
            return (1 - self.l)*poss_val - self.l*reduct_val/self.__dim
        elif self.method == 'dominance':
            return ArrayPair(reduct_val, poss_val)
        else:
            raise ValueError("%s is not an allowed value for 'method' parameter" % self.method)



    def __evaluate_poss(self, temp_y):
        poss_values = np.zeros(self.__n_instances)
        for i in range(self.__n_instances):
            poss_values[i] = self.y[i][temp_y[i]]

        if self.method == 'conservative' or self.method == 'lambda':
            return np.amin(poss_values)
        elif self.method == 'dominance':
            return poss_values
        else:
            raise ValueError("%s is not an allowed value for 'method' parameter" % self.method)



    def __evaluate_reduct(self, temp_X, temp_y):
        nn = None
        if self.discrete:
            nn = RadiusNeighborsClassifier(radius = 0.0,
                                          metric = 'hamming')
        elif self.neighborhood == 'delta':
            nn = RadiusNeighborsClassifier(radius=self.radius,
                                           metric=self.metric)
        elif self.neighborhood == 'nearest':
            nn = KNeighborsClassifier(n_neighbors = self.n_neighbors,
                                      metric = self.metric)
        else:
            raise ValueError("%s is not an allowed value for 'neighborhood' parameter" % self.neighborhood)
        
        if temp_X.shape[1] == 0:
            return np.inf

        nn_all = clone(nn)
        nn.fit(temp_X, temp_y)
        nn_all.fit(self.X, temp_y)

        accuracy = accuracy_score(nn_all.predict(self.X), nn.predict(temp_X))
        check = (accuracy >= (1 - self.epsilon))
        
        return temp_X.shape[1] if check else np.inf
        
    def fit(self, X, y):
        """
        Fit the GeneticRoughSetSelector model
        """
        state = np.random.get_state()
        if not (self.random_state is None):
            np.random.seed(self.random_state)


        self.X = X
        self.y = np.array(y)
        self.__basic_type = y[0].basic_type()
        self.__dim = X.shape[1]
        self.__n_instances = X.shape[0]
        self.n_classes_ = y[0].n_classes
        self._initialize()

        self.__fitnesses = np.empty(self.population_size, dtype=FitnessVal)
        for i in range(self.n_iters):
            self.__best_value = self.__compute_fitness(0)
            self.__fitnesses[0] = self.__best_value
            self.best_features_ = [self.__population_features[0,:]]
            self.best_targets_ = [self.__population_targets[0]]

            
            for p in range(1,self.population_size):
                self.__fitnesses[p] = self.__compute_fitness(p)
                if self.__fitnesses[p] == self.__best_value:
                    self.best_features_.append(self.__population_features[p,:])
                    self.best_targets_.append(self.__population_targets[p])
                elif self.__fitnesses[p] > self.__best_value:
                    self.__best_value = self.__fitnesses[p]
                    self.best_features_ = [self.__population_features[p,:]]
                    self.best_targets_ = [self.__population_targets[p]]

            self.__breed()

        if not (self.random_state is None):
            np.random.set_state(state)

        self.best_features_ = np.array(self.best_features_)
        self.best_targets_ = np.array(self.best_targets_)
        _, idx = np.unique(self.best_features_, axis=0, return_index=True)
        self.best_features_ = self.best_features_[idx]
        self.best_targets_ = self.best_targets_[idx]

        return self

    def transform(self, X, y=None):
        """
        Transform the data (only X, y is ignored) selecting a reduct at random
        """
        state = np.random.get_state()
        if not (self.random_state is None):
            np.random.seed(self.random_state)
            
        feature_set = np.random.choice(len(self.best_features_))

        if not (self.random_state is None):
            np.random.set_state(state)

        return X[:, self.best_features_[feature_set]]

    def fit_transform(self, X, y):
        """
        Fit and then transform data
        """
        self.fit(X,y)
        return self.transform(X, y)





from abc import ABC, abstractmethod
import numpy as np

class GenericWeakLabel(ABC):
    @abstractmethod
    def sample_value(self):
        pass

    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def basic_type(self):
        pass


class FuzzyLabel(ABC):
    @abstractmethod
    def __getitem__(self, val):
        pass

    @abstractmethod
    def get_cut(self, alpha):
        pass


class ContinuousWeakLabel(GenericWeakLabel):
    def basic_type(self):
        return np.float32


class IntervalFuzzyLabel(ContinuousWeakLabel,FuzzyLabel):
    pass

class IntervalLabel(IntervalFuzzyLabel):
    def __init__(self, lower, upper):
        if lower > upper:
            raise ValueError("Lower bound cannot be greater than upper bound")
        self.lower = lower
        self.upper = upper

    def sample_value(self):
        return np.random.random()*(self.upper - self.lower) + self.lower

    def __eq__(self, other):
        if isinstance(other, IntervalLabel):
            return (self.lower == other.lower) and (self.upper == other.upper)
        else:
            return False

    def __getitem__(self, val):
        return 1 if (val >= self.lower and val <= self.upper) else 0

    def get_cut(self, alpha):
        if alpha < 0 or alpha > 1:
            raise ValueError("Alpha should be between 0 and 1")
        else:
            return self

    def __str__(self):
        return "[%f, %f]" % (self.lower, self.upper)


class DiscreteWeakLabel(GenericWeakLabel):
    def basic_type(self):
        return np.int32
    
    @abstractmethod
    def to_probs(self):
        pass

class DiscreteFuzzyLabel(DiscreteWeakLabel, FuzzyLabel):
    def __init__(self, classes, n_classes):
        self.n_classes = n_classes
        if isinstance(classes, np.ndarray) and classes.ndim == 1:
            if classes.shape[0] > self.n_classes:
                raise ValueError("Unexpected number of classes")
            self.classes = classes
        elif isinstance(classes, dict):
            self.classes = self.__to_array_encoding(classes)
        elif isinstance(classes, list):
            self.classes = np.zeros(self.n_classes)
            for i in classes:
                self.classes[i] = 1
        else:
            raise ValueError("Format not supported")

    def to_probs(self):
        probs= np.zeros(self.n_classes)
        values = np.unique(self.classes)
        sorted_values = np.sort(values)[::-1]
        if sorted_values[-1] != 0.0:
            sorted_values = np. append(sorted_values, 0.0)
        for j in range(len(sorted_values) - 1):
            val = sorted_values[j]
            idx = np.where(self.classes >= val)
            assign = (val - sorted_values[j+1])/len(idx[0])
            probs[idx] += assign
        return probs

    def sample_value(self):
        probs= self.to_probs()
        return np.random.choice(range(self.n_classes), p=probs)

    def __to_array_encoding(self, classes):
        out = np.zeros(self.n_classes)
        for i in classes:
          if out[i] > 1:
            raise ValueError("Values greater than 1 are not allowed")
          out[i] = classes[i]
        return out

    def __getitem__(self, val):
        return self.classes[val]

    def get_cut(self, alpha):
        if alpha < 0 or alpha > 1:
            raise ValueError("Alpha should be between 0 and 1")
        return DiscreteSetLabel((self.classes >= alpha) * 1.0, self.n_classes)

    def __eq__(self, other):
        if isinstance(other, DiscreteWeakLabel):
            return np.array_equal(self.classes, other.classes)
        else:
            return False

    def __str__(self):
        return str(self.classes)

    def is_agnostic(self):
        return np.all(np.equal(self.classes, np.ones_like(self.classes)))


class DiscreteSetLabel(DiscreteFuzzyLabel):

    def __init__(self, classes, n_classes):
      super().__init__(classes, n_classes)
      for i in self.classes:
        if i not in [0.0, 1.0]:
          raise ValueError("Only values equal to 1 are allowed")
    
    def sample_value(self):
        return np.random.choice(range(self.n_classes), p=self.classes/np.sum(self.classes))


from functools import partial

import numpy as np
from keras.optimizer_v2.gradient_descent import SGD
from sklearn.base import BaseEstimator, ClassifierMixin
import tensorflow as tf
# from tensorflow.keras.wrappers.scikit_learn import KerasClassifier, BaseWrapper
from keras.metrics import accuracy
from keras.wrappers.scikit_learn import BaseWrapper
from keras.models import Sequential
from keras.layers import Input, Dense
from keras.regularizers import l2
from keras.losses import kullback_leibler_divergence
import keras.backend as keras_backend
from sklearn.preprocessing import OneHotEncoder


class LabelRelaxationLoss(object):
    """
    Wrapper object that provides the label relaxation loss as described in "From Label Smoothing to Label Relaxation" by
    Julian Lienen and Eyke Huellermeier, AAAI 2021.

    Parameters
    ----------
    :param alpha: Imprecisiation degree alpha of the label relaxation loss
    :type alpha: float, default=0.1
    """

    def __init__(self, alpha: float = 0.1):
        super().__init__()

        self._alpha = alpha

    def loss(self, y_true, y_pred):
        if self._alpha is None:
            y_true, alphas = y_true
        else:
            alphas = self._alpha

        y_pred = y_pred / tf.reduce_sum(y_pred, axis=-1, keepdims=True)
        y_pred = tf.clip_by_value(y_pred, keras_backend.epsilon(), 1. - keras_backend.epsilon())

        sum_y_hat_prime = tf.reduce_sum((1. - y_true) * y_pred, axis=-1)
        y_pred_hat = alphas * y_pred / (tf.expand_dims(sum_y_hat_prime, axis=-1) + keras_backend.epsilon())
        y_true_credal = tf.where(tf.greater(y_true, 0.1), 1. - alphas, y_pred_hat)

        divergence = kullback_leibler_divergence(y_true_credal, y_pred)

        preds = tf.reduce_sum(y_pred * y_true, axis=-1)

        return tf.where(tf.greater_equal(preds, 1. - alphas), tf.zeros_like(divergence),
                        divergence)


def accuracy_metric(y_true, y_pred, tuple_target=False, n_classes=10):
    if tuple_target:
        targets = tf.reshape(y_true[0], (-1, n_classes))
    else:
        targets = y_true

    # Argmax as we face probability distributions
    y_pred = tf.argmax(y_pred, axis=-1)
    targets = tf.argmax(targets, axis=-1)
    return accuracy(targets, y_pred)


class LabelRelaxationNNClassifier(BaseEstimator, ClassifierMixin):
    """
    Simple MLP-based classifier using the label relaxation loss as optimization criterion.

    Parameters
    ----------
    :param lr_alpha: Imprecisiation degree alpha of the label relaxation loss
    :type lr_alpha: float, default=0.1

    :param hidden_layer_sizes: Tuple consisting of the individual hidden layer sizes used by the underlying NN model
    :type hidden_layer_sizes: tuple, default=(100,)

    :param activation: Activation function applied to the hidden layers' activations
    :type activation: str, default 'relu'

    :param l2_penalty: L2 norm regularization parameter applied to all neural network layers
    :type l2_penalty: float, default=1e-4

    :param learning_rate: Learning rate used by the SGD optimizer
    :type learning_rate: float, default=1e-3

    :param momentum: Momentum used by the SGD optimizer
    :type momentum: float, default=0.0

    :param epochs: Number of training epochs
    :type epochs: int, default=100

    :param batch_size: Batch size for training
    :type batch_size: int, default=None

    :param provide_alphas: Indicator whether we consider tuples as targets consisting of the classes and their
        imprecisiation
    :type provide_alphas: bool, default=False

    :param n_classes: Number of classes in case we want to be certain about the dimensionality of the one-hot encoding
    :type n_classes: int default=None
    """

    def __init__(self, lr_alpha: float = 0.1, hidden_layer_sizes: tuple = (100,), activation: str = "relu",
                 l2_penalty: float = 1e-4, learning_rate: float = 1e-3, momentum: float = 0.0, epochs: int = 100,
                 batch_size: int = None, provide_alphas: bool = False, n_classes: int = None):
        super().__init__()

        self._lr_loss = LabelRelaxationLoss(lr_alpha)
        self._hidden_layer_sizes = hidden_layer_sizes
        self._hidden_layer_activation = activation
        self._l2_penalty = l2_penalty
        self._learning_rate = learning_rate
        self._momentum = momentum
        self._epochs = epochs
        self._batch_size = batch_size
        self.n_classes = n_classes

        self._internal_model = None

        self.provide_alphas = provide_alphas

    def _one_hot_encoding(self, targets):
        if self.n_classes is None:
            enc = OneHotEncoder(sparse=False)
            return enc.fit_transform(targets.reshape(-1, 1))
        else:
            # targets = np.take(np.eye(self.n_classes), targets, axis=0)
            one_hot_encoded = np.zeros((len(targets), self.n_classes))
            one_hot_encoded[np.arange(len(targets)), targets] = 1
            return one_hot_encoded

    def fit(self, X, y):
        """
        Fits the label relaxation model. The targets y are one-hot encoded in case a simple list is provided.
        """
        input_dim = X.shape[1]

        targets = y if not self.provide_alphas else y[0]
        if len(targets.shape) < 2:
            targets = self._one_hot_encoding(targets)
        self.n_classes = targets.shape[1]
        if self.provide_alphas:
            y = (targets, y[1])
        else:
            y = targets

        def _create_model():
            model = Sequential()
            model.add(Input(input_dim))
            for hl_size in self._hidden_layer_sizes:
                model.add(
                    Dense(hl_size, activation=self._hidden_layer_activation, kernel_regularizer=l2(self._l2_penalty)))
            model.add(Dense(self.n_classes, activation="softmax", kernel_regularizer=l2(self._l2_penalty)))

            optimizer = SGD(lr=self._learning_rate, momentum=self._momentum)
            acc_metric = partial(accuracy_metric, tuple_target=self.provide_alphas, n_classes=self.n_classes)
            acc_metric.__name__ = "accuracy"
            model.compile(loss=self._lr_loss.loss, optimizer=optimizer,
                          metrics=[acc_metric])
            return model

        self._internal_model = BaseWrapper(build_fn=_create_model, epochs=self._epochs, batch_size=self._batch_size)
        self._internal_model.fit(X, y)
        return self

    def predict(self, X):
        assert self._internal_model is not None, "Model needs to be fit before prediction."

        return self._internal_model.model.predict(X)

    def predict_proba(self, X):
        # As last layer is a softmax layer
        return self.predict(X)



import pytest
from sklearn.datasets import load_iris
from sklearn.neighbors import KNeighborsClassifier
from ...utils import DiscreteRandomSmoother, DiscreteEstimatorSmoother

@pytest.fixture
def dataset():
    X, y = load_iris(return_X_y=True)
    return (X,y)

def test_simple_random(dataset):
    X, y = dataset[0], dataset[1]
    DiscreteRandomSmoother().fit(X,y)
    DiscreteRandomSmoother().fit_transform(X,y)
    assert True

def test_simple_estimator(dataset):
    X, y = dataset[0], dataset[1]
    est = KNeighborsClassifier()
    DiscreteEstimatorSmoother(estimator=est).fit(X,y)
    DiscreteEstimatorSmoother(estimator=est).fit_transform(X,y)
    assert True



import pytest
from sklearn.datasets import load_iris
from ...classification._relaxation import LabelRelaxationNNClassifier


@pytest.fixture
def dataset():
    X, y = load_iris(return_X_y=True)
    return X, y


def test_label_relaxation(dataset):
    """
    Simple label relaxation test.
    """

    X, y_true = dataset[0], dataset[1]
    clf = LabelRelaxationNNClassifier(lr_alpha=0.1, hidden_layer_sizes=(128,), activation="relu",
                                      l2_penalty=1e-8, learning_rate=1e-1, momentum=0.0, epochs=100,
                                      batch_size=128)
    clf.fit(X, y_true)
    _ = clf.predict(X)
    _ = clf.predict_proba(X)
    assert True



from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.tree import ExtraTreeClassifier
from sklearn.preprocessing import normalize
from sklearn.utils import resample
import numpy as np

class RRLClassifier(BaseEstimator, ClassifierMixin):
  '''
    A class to perform classification for weakly supervised data, based on the RRL algorithm [1].
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel

    Parameters
    ----------
    :param estimator: Base estimator objects to be fitted. Should support predict and predict_proba
    :type estimator: estimator class, default=ExtraTreeClassifier

    :param n_estimators: The number of trees to be fitted
    :type n_estimators: int, default=100

    :param resample: Whether to perform bootstrapping or not
    :type resample: bool, default=False
        
    :param random_state: Random seed
    :type random_state: int, default=None

    Attributes
    ----------

    :ivar classifiers: The collection of fitted estimators
    :vartype classifiers: list of estimators

    :ivar __ys: The collection of sampled target labels. Each ndarray in ys has the same shape as y
    :vartype __ys: list of ndarrays
    
    :ivar __Xs: The collection of bootstrapped datasets. Each ndarray in Xs has the same shape as X. If resample=False, then Xs is empty.
    :vartype __Xs: list of ndarrays
    
    :ivar __n_classes: The number of unique classes in y
    :vartype __n_classes: int

    :ivar __classes: The unique classes in y
    :vartype __classes: list of int
    '''
  def __init__(self, estimator=ExtraTreeClassifier(), n_estimators=100, resample=False, random_state=None):
    self.n_estimators = n_estimators
    self.random_state = random_state
    self.estimator = estimator
    self.resample = resample

  def fit(self, X, y):
    """
        Fit the RRLClassifier model
    """
    state = np.random.get_state()
    if not (self.random_state is None):
      np.random.seed(self.random_state)

    self.__X = X
    self.__y = np.array(y)

    self.__n_classes = self.__y[0].n_classes
    self.__classes = range(self.__n_classes)
    
    self.__classifiers = [ clone(self.estimator) for i in range(self.n_estimators)]
    
    self.__ys = []
    self.__Xs = []
    for i in range(self.n_estimators):
      temp_ys = self.__sample_labels()
      seed = np.random.randint(np.iinfo('int32').max)
      self.__classifiers[i].set_params(**{'random_state': seed})

      if self.resample:
        seed = np.random.randint(np.iinfo('int32').max)
        temp_Xs, temp_ys = resample(self.__X, temp_ys, random_state = seed)
        self.__Xs.append(temp_Xs)
        self.__ys.append(temp_ys)
        self.__classifiers[i].fit(self.__Xs[i], self.__ys[i])
      else:
        self.__ys.append(temp_ys)
        self.__classifiers[i].fit(self.__X, self.__ys[i])

    if not (self.random_state is None):
      np.random.set_state(state)
    return self

  def __sample_labels(self):
    y_res = np.empty(len(self.__y))
    for i in range(self.__X.shape[0]):
      y_res[i] = self.__y[i].sample_value()
    return y_res
  
  def predict(self, X):
    """
        Returns predictions for the given X
    """
    output = np.zeros((X.shape[0], self.__n_classes))
    sorter = np.argsort(self.__classes)

    for i in range(self.n_estimators):
      indices = sorter[np.searchsorted(self.__classes, self.__classifiers[i].classes_, sorter=sorter)]
      output[:, indices] += self.__classifiers[i].predict_proba(X)
    return np.argmax(output, axis= 1)

  def predict_proba(self, X):
    """
        Returns probability distributions for the given X
    """
    output = np.zeros((X.shape[0], self.__n_classes))
    sorter = np.argsort(self.__classes)
    for i in range(self.n_estimators):
      indices = sorter[np.searchsorted(self.__classes, self.__classifiers[i].classes_, sorter=sorter)]
      output[:, indices] += self.__classifiers[i].predict_proba(X)
    return normalize(output, axis=1, norm='l1')

  def fit_predict(self, X, y):
    self.fit(X,y)
    return self.predict(X)


import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.base import BaseEstimator, ClassifierMixin

class WeaklySupervisedKNeighborsClassifier(BaseEstimator, ClassifierMixin):
    '''
    A class to perform classification for weakly supervised data, based on k-nearest neighbors.
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel

    Parameters
    ----------
    :param k: The number of neighbors
    :type k: int, default=3

    :param metric: The metric for neighbors queries
    :type metric: str or callable, default 'minkowski'

    Attributes
    ----------
    :ivar __n_classes: The number of unique classes in y
    :vartype __n_classes: int

    :ivar __classes: The unique classes in y
    :vartype __classes: list of int
    '''
    def __init__(self, k=3, metric='minkowski'):
        self.k = k
        self.metric = metric
        
    def fit(self, X, y):
        """
        Fit the WeaklySupervisedKNeighborsClassifier model
        """
        self.__X = X
        self.__y = np.zeros((len(y), y[0].n_classes))
        for i in range(len(y)):
            self.__y[i] = y[i].classes

        self.__n_classes = y[0].n_classes
        self.__classes = range(self.__n_classes)
        
        self.__tree = NearestNeighbors(metric=self.metric, n_neighbors=self.k)
        self.__tree.fit(self.__X)
        return self

    
    def predict(self, X):
        """
        Returns predictions for the given X
        """
        y_pred = np.zeros(X.shape[0])
        _, indices = self.__tree.kneighbors(X)
        for i in range(len(indices)):
            y_pred[i] = np.argmax(np.add.reduce(self.__y[indices[i]]))
        return y_pred
    
    def predict_proba(self, X):
        """
        Returns probability distributions for the given X
        """
        y_pred = np.zeros((X.shape[0], self.__n_classes))
        _, indices = self.__tree.kneighbors(X)
        for i in range(len(indices)):
            y_pred[i, :] = np.add.reduce(self.__y[indices[i]])
            y_pred[i, :] /= np.sum(y_pred[i,:])
        return y_pred







class WeaklySupervisedRadiusClassifier(BaseEstimator, ClassifierMixin):
    '''
    A class to perform classification for weakly supervised data, based on radius neighbors.
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel

    Parameters
    ----------
    :param radius: The size of the radius
    :type radius: float, default=1.0

    :param metric: The metric for neighbors queries
    :type metric: str or callable, default 'minkowski'

    Attributes
    ----------
    :ivar y: A copy of the input y
    :vartype y: ndarray

    :ivar n_classes: The number of unique classes in y
    :vartype n_classes: int
    '''
    def __init__(self, radius=1.0, metric='minkowski'):
        self.radius = radius
        self.metric = metric
        
    def fit(self, X, y):
        """
        Fit the WeaklySupervisedRadiusClassifier model
        """
        self.__X = X
        self.__y = np.zeros((len(y), y[0].n_classes))
        for i in range(len(y)):
            self.__y[i] = y[i].classes
        
        self.__tree = NearestNeighbors(metric=self.metric, radius=self.radius)
        self.__tree.fit(self.__X)
        self.__n_classes = y[0].n_classes
        return self

    
    def predict(self, X):
        """
        Returns predictions for the given X
        """
        y_pred = np.zeros(X.shape[0])
        _, indices = self.__tree.radius_neighbors(X)
        for i in range(len(indices)):
            y_pred[i] = np.argmax(np.add.reduce(self.__y[indices[i]]))
        return y_pred
    
    def predict_proba(self, X):
        """
        Returns probability distributions for the given X
        """
        y_pred = np.zeros((X.shape[0], self.__n_classes))
        _, indices = self.__tree.radius_neighbors(X)
        for i in range(len(indices)):
            y_pred[i, :] = np.add.reduce(self.__y[indices[i]])
            y_pred[i, :] /= np.sum(y_pred[i,:])
        return y_pred


from ._ensemble import RRLClassifier
from ._neighbors import WeaklySupervisedKNeighborsClassifier, WeaklySupervisedRadiusClassifier
from ._grm import GRMLinearClassifier
from ._pseudolabels import PseudoLabelsClassifier, CSSLClassifier
from ._relaxation import LabelRelaxationNNClassifier

__all__ = [
    "RRLClassifier",
    "WeaklySupervisedKNeighborsClassifier",
    "WeaklySupervisedRadiusClassifier",
    "GRMLinearClassifier",
    "PseudoLabelsClassifier",
    "CSSLClassifier",
    "LabelRelaxationNNClassifier"
]


from setuptools import setup, find_packages

setup(
   name='scikit-weak',
   version='0.2.1',
   author='Andrea Campagner',
   python_requires=">3.8.0",
   author_email='a.campagner@campus.unimib.it',
   packages=find_packages(include=['scikit_weak', 'scikit_weak.*']),
   url='https://pypi.org/project/scikit-weak/',
   license='LICENSE.txt',
   description='A package featuring utilities and algorithms for weakly supervised ML.',
   long_description=open('README.md').read(),
   long_description_content_type='text/markdown',
   install_requires=[
       "scikit-learn>=1.0.0",
       "numpy>=1.19.2",
       "scipy>=1.3.2",
       "tensorflow>=2.4.0",
       "keras>=2.4.0",
       "pytest"
   ],
)


from ._representations import (GenericWeakLabel, ContinuousWeakLabel, DiscreteWeakLabel, FuzzyLabel,
                              DiscreteFuzzyLabel, DiscreteSetLabel, IntervalFuzzyLabel, IntervalLabel)

__all__ = [
    "GenericWeakLabel",
    "ContinuousWeakLabel",
    "DiscreteWeakLabel",
    "FuzzyLabel",
    "DiscreteFuzzyLabel",
    "DiscreteSetLabel",
    "IntervalFuzzyLabel",
    "IntervalLabel"
]


from scipy.linalg import eigh
from ..classification import WeaklySupervisedKNeighborsClassifier
from ..data_representation import DiscreteFuzzyLabel
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin

class DELIN(BaseEstimator, TransformerMixin, ClassifierMixin):
    '''
    A class to perform classification and dimensionality reduction for weakly supervised data,
    based on the DELIN algorithm [1]. The original method has been slighlty modified by using
    SVD in the computation of the inverse, so as to avoid issues when the data matrix is singular.
    The y input to the fit method should be given as an iterable of DiscreteWeakLabel

    Parameters
    ----------
    :param k: The number of neighbors
    :type k: int, default=3
    
    :param d: The number of dimensions to be kept after reduction. If int, then the exact number of features to be kept.
    If float, the percentage of the total number of dimensions.
    :type d: int or float, default = 2
        
    :param iters: The number of iterations to be performed
    :type iters: int, default = 10

    Attributes
    ----------
    :ivar y: If y is in prob format, then target is a copy of y. Otherwise it is y in prob format
    :vartype y: ndarray
            
    :ivar clf: A WeaklyKNeighborsClassifier classifier to be used during fitting of the algorithm
    :vartype clf: WeaklyKNeighborsClassifier object

    :ivar vr: A square ndarray with the same dim as X.shape[1]. Used to perform dimensionality reduction
    :vartype vr: ndarray
            
    :ivar n_classes: The number of unique classes in y
    :vartype n_classes: int

    :ivar classes_: The unique classes in y
    :vartype classes: ndarray
    '''
    def __init__(self, k=3, d=2, n_iters = 10):
        self.k = k
        self.d = d
        self.n_iters = n_iters
        
    def fit(self, X, y):
        """
        Fit the DELIN model
        """
        self.__X = X
        self.__y = np.array(y)
        
        
        self.__clf = WeaklySupervisedKNeighborsClassifier(k=self.k)
        self.__n_classes = self.__y[0].n_classes

        if type(self.d) == float:
            self.d = int(self.d*X.shape[1])
        
        self.__y_probs = np.empty((self.__X.shape[0], self.__n_classes))

        for i in range(len(self.__y)):
            self.__y_probs[i] = self.__y[i].to_probs()
        self.__y_prime = self.__y_probs.copy()

        for it in range(self.n_iters):
            self.__mu = np.mean(self.__X, axis = 0)
            self.__Mu = np.zeros((self.__n_classes, self.__X.shape[1]))

            for j in range(self.__n_classes):
                self.__Mu[j, :] = np.mean(np.multiply(self.__X, np.transpose([self.__y_probs[:,j]]) ), axis=0)

            X_hat = self.__X - self.__mu
            S_t = X_hat.T.dot(X_hat)
            C = np.zeros((self.__n_classes, self.__n_classes))
            for i in range(self.__n_classes):
                C[i,i] = np.sum(self.__y_probs[:,i])
            S_b = X_hat.T.dot(self.__y_prime).dot(np.linalg.pinv(C)).dot(self.__y_prime.T).dot(X_hat)
            S_w = S_t - S_b

            Mat = np.linalg.pinv(S_w).dot(S_b)
            _, self.__vr= eigh(Mat)
            
            X_prime = self.__X.dot(self.__vr[:,-self.d:])
            y_temp = np.empty(self.__y_prime.shape[0], DiscreteFuzzyLabel)

            for i in range(self.__y_prime.shape[0]):
                y_temp[i] = DiscreteFuzzyLabel(self.__y_prime[i], self.__n_classes)

            self.__y_prime = self.__clf.fit(X_prime, y_temp).predict_proba(X_prime)
            

            self.__y_prime = self.__y_prime*self.__y_probs
            
            for i in range(self.__y_prime.shape[0]):
                check = True
                for j in range(self.__y_prime.shape[1]):
                    if self.__y_prime[i,j] > 0:
                        check = False
                        break
                if check:
                    self.__y_prime[i] = self.__y_probs[i]

            self.__y_prime = self.__y_prime/self.__y_prime.sum(axis=1, keepdims=True)

        return self

    def fit_predict(self, X, y):
        self.fit(X, y)
        return self.predict(X)
    
    def predict(self, X):
        """
        Returns predictions for the given X
        """
        X_prime = X.dot(self.__vr[:,-self.d:])
        return self.__clf.predict(X_prime)
    
    def predict_proba(self, X):
        """
        Returns probability distributions for the given X
        """
        X_prime = X.dot(self.__vr[:,-self.d:])
        return self.__clf.predict_proba(X_prime)

    def transform(self, X, y=None):
        X_prime = X.dot(self.__vr[:,-self.d:])
        return X_prime

    def fit_transform(self, X, y):
        self.fit(X,y)
        return self.transform(X)



import numpy as np
import scipy.stats as stats

def to_probs(ys, uniform=False):
    if ys.ndim < 2:
        values = np.unique(np.add.reduce(ys))
        probs = np.zeros((len(ys),len(values)))
        for i in range(len(ys)):
            for val in ys[i]:
                idx = values[values == val]
                probs[i,idx] += 1
            probs[i,:] /= np.sum(probs[i,:])
        return probs
    else:
        probs = np.zeros(ys.shape)
        if uniform:
            probs = ys/np.sum(ys, axis=1)[:, np.newaxis]
        else:
            for i in range(ys.shape[0]):
                values = np.unique(ys[i])
                sorted_values = np.sort(values)[::-1]
                if sorted_values[-1] != 0.0:
                    sorted_values = np. append(sorted_values, 0.0)
                for j in range(len(sorted_values) - 1):
                    val = sorted_values[j]
                    idx = np.where(ys[i,:] >= val)
                    assign = (val - sorted_values[j+1])/len(idx[0])
                    probs[i, idx] += assign
        return probs



def prob_format(ys):
    if ys.ndim < 2:
        return False
    val_low = 1 - np.finfo(np.float32).eps
    val_upp = 1 + np.finfo(np.float32).eps
    vals = np.sum(ys, axis=1)
    if np.all(( vals >= val_low) & (vals <= val_upp) ):
        return True
    return False


def oau_entropy(orthop, n_classes):
    orthop_copy = np.copy(orthop)
    probs = np.zeros([n_classes, len(orthop)])
    i = 0
    for elem in orthop:
        for cl in elem:
            probs[cl, i] = 1.0
        i += 1

    tots = np.sum(probs, axis=1)
    indices = np.argsort(tots)[::-1]
    for i in indices:
        for j in range(len(orthop)):
            if i in orthop_copy[j]:
                orthop_copy[j] = [i]
    orthop_copy = np.array([t[0] for t in orthop_copy])
    probs = np.zeros([n_classes])
    for elem in orthop_copy:
        probs[elem] += 1.0
    probs /= sum(probs)
    return stats.entropy(probs, base=2)

def bet_entropy(orthop, n_classes):
        probs = np.zeros([n_classes])
        unif = np.ones([n_classes])
        for elem in orthop:
            for cl in elem:
                probs[cl] += 1.0/len(elem)
        probs /= sum(probs)
        unif = unif * (probs > 0)
        unif /= sum(unif)
        num =  stats.entropy(probs, base=2)
        den = stats.entropy(unif, base=2)
        h = 0 if num == 0 else num/den
        return h


import tensorflow as tf

@tf.function
def fuzzy_cross_entropy(y_true, y_pred):
    return fuzzy_loss(y_true, y_pred,
                      tf.keras.losses.SparseCategoricalCrossentropy(reduction=tf.keras.losses.Reduction.NONE))
@tf.function
def fuzzy_hinge(y_true, y_pred):
    return fuzzy_loss(y_true, y_pred,
                      tf.keras.losses.CategoricalHinge(reduction=tf.keras.losses.Reduction.NONE))

@tf.function
def fuzzy_loss(y_true, y_pred, loss_function):
    i = tf.constant(0)
    acc_outer = tf.constant(0.0)
    
    def outer_loop(i, acc_outer):
        y, _ = tf.unique(y_true[i])
        cuts = tf.sort(y, direction="DESCENDING")
        
        n_cuts = cuts.shape[0]
        j = tf.constant(0)
        acc_cut = tf.constant(0.0)
        
        def cut_loop(j, acc_cut):
            indices = tf.where(y_true[i] >= cuts[j])
            cut_size = indices.shape[0]
            
            k = tf.constant(0)
            acc_inner = tf.constant(0.0)
            
            def inner_loop(k, acc_inner):
                val = tf.reshape(tf.tile(y_pred[i,:], [cut_size]), (cut_size, y_true.shape[1]))
                loss = tf.reduce_min(loss_function(indices, val))
                return (tf.add(k,1), tf.add(acc_inner,loss))
            
            c_inner = lambda k, _: tf.less(k, cut_size)
            b_inner = lambda k, acc_inner: inner_loop(k, acc_inner)
            _, acc_inner = tf.while_loop(c_inner, b_inner, [k, acc_inner])
            return (tf.add(j, 1), tf.divide(tf.multiply(cuts[j] - cuts[j+1], acc_inner), cut_size))
            
        
        c_cut = lambda j, _ : tf.less(j, n_cuts-1)
        b_cut = lambda j, acc_cut: cut_loop(j, acc_cut)
        _, acc_cut = tf.while_loop(c_cut, b_cut, [j,acc_cut])
        
        return (tf.add(i,1), tf.add(acc_outer, acc_cut))
    
    c_outer = lambda i, _ : tf.less(i, y_pred.shape[0])
    b_outer = lambda i, acc_outer: outer_loop(i, acc_outer)
    _, acc_outer = tf.while_loop(c_outer, b_outer, [i, acc_outer])
    return acc_outer